import time
import cv2
import numpy as np
import tensorflow as tf
from pathlib import Path
from PySide6.QtCore import Slot, QStandardPaths
from PySide6.QtWidgets import QMainWindow, QFileDialog, QMessageBox
from PySide6.QtGui import QPixmap
from covid19.ui._ui_client import Ui_Client
from covid19.models import ResNet50, COVIDNet
from covid19.explainers import GradCAM, IG


def _get_size(ui):
    size = ui.maximumSize()
    return size.height(), size.width()


def numpy_to_pixmap(image, size):
    tmp_file = '/tmp/covid_{}.png'.format(time.time())
    image = cv2.resize(image, size)
    cv2.imwrite(tmp_file, image)
    image = QPixmap(tmp_file)
    tmp_file = Path(tmp_file)
    tmp_file.unlink()
    return image


class Client(QMainWindow):
    def __init__(self, n_classes, models_path, class_labels):
        super().__init__()
        self.n_classes = n_classes
        self.models_path = Path(models_path)
        self.class_labels = {l: c for (c, l) in class_labels.items()}
        self._model = None
        self._model_loaded = False
        self._explainer = None
        self._input = None
        self._input_predicted = True

        self.ui = Ui_Client()
        self.ui.setupUi(self)
        self.ui.model.currentTextChanged.connect(self._refresh_explainer)
        self.ui.model.currentTextChanged.connect(self._refresh_predict)
        self.ui.explainer.currentTextChanged.connect(self._refresh_predict)
        self.ui.predict.clicked.connect(self._refresh_predict)
        self.ui.select_image.clicked.connect(self._refresh_predict)

        self._input_size = _get_size(self.ui.input)
        self._explanation_size = _get_size(self.ui.explanation)
        self._refresh_model()
        self.ui.predict.setEnabled(False)

    @Slot()
    def on_predict_clicked(self):
        # forward
        print('Predicting... ', end='', flush=True)
        input_ = cv2.resize(self._input, self._model.image_shape[0:2])
        prediction, confidence, explanation = self._explainer.explain(input_)
        explanation = (explanation * 255).astype(np.uint8)              # [0,1] -> [0,255]
        explanation = cv2.cvtColor(explanation, cv2.COLOR_BGR2RGB)      # invert colors
        print('done')

        # update GUI
        self.ui.prediction.setText(self.class_labels[prediction])
        self.ui.confidence.setText('{:.2f}%'.format(confidence * 100))
        self.ui.explanation.setPixmap(numpy_to_pixmap(explanation, self._explanation_size))
        self._input_predicted = True

    @Slot()
    def on_select_image_clicked(self):
        print('Selecting image... ', end='', flush=True)
        filename, _ = QFileDialog.getOpenFileName(
            parent=self,
            caption='Select chest X-ray image',
            dir=QStandardPaths.standardLocations(QStandardPaths.PicturesLocation).pop(),
            filter='Images (*.png *.jpg *.jpeg)'
        )
        if filename != '':
            self._input = cv2.imread(filename)
            self.ui.input.setPixmap(numpy_to_pixmap(self._input, self._input_size))
            self.ui.prediction.clear()
            self.ui.confidence.clear()
            self.ui.explanation.setPixmap(':/images/default.png')
            self._input_predicted = False
            self.ui.input_state.setPixmap(':/images/ok.png')
            print(filename)
        else:
            print('canceled')

    @Slot(str)
    def on_model_currentTextChanged(self, model):
        print('Loading {}... '.format(model), end='', flush=True)
        if model == 'ResNet50':
            self._model = ResNet50(self.n_classes, weights=None)
            model_path = self.models_path / 'resnet50'
        elif model == 'COVID-Net':
            self._model = COVIDNet(self.n_classes, weights=None)
            model_path = self.models_path / 'covidnet'
        else:
            raise ValueError('Invalid model {}'.format(model))
        try:
            self._model.load_weights(model_path)
            self._model_loaded = True
            self._input_predicted = False
            self.ui.model_state.setPixmap(':/images/ok.png')
            print('done')
        except tf.errors.NotFoundError:
            self._model_loaded = False
            self.ui.model_state.setPixmap(':/images/error.png')
            QMessageBox.critical(
                self,
                'Model not found',
                'Failed to load model, check that the models directory contains the model.'
            )
            print('failed')

    @Slot(str)
    def on_explainer_currentTextChanged(self, explainer):
        print('Preparing {}... '.format(explainer), end='', flush=True)
        if explainer == 'Grad-CAM':
            self._explainer = GradCAM(self._model)
        elif explainer == 'Integrated Gradients':
            self._explainer = IG(self._model)
        else:
            raise ValueError('Invalid explainer {}'.format(explainer))
        self.ui.explainer_state.setPixmap(':/images/ok.png')
        self._input_predicted = False
        print('done')

    @Slot()
    def _refresh_model(self):
        self.ui.model.currentTextChanged.emit(self.ui.model.currentText())

    @Slot()
    def _refresh_explainer(self):
        self.ui.explainer.currentTextChanged.emit(self.ui.explainer.currentText())

    @Slot()
    def _refresh_predict(self):
        enabled = self._model_loaded and not self._input_predicted
        self.ui.predict.setEnabled(enabled)



#!/usr/bin/env python3

import argparse
from utils import get_model, get_callbacks, get_class_weights, IMAGE_SIZE
from pathlib import Path
from covid19.datasets import image_dataset_from_directory
from covid19.metrics import plot_learning_curves
from tensorflow.keras.losses import CategoricalCrossentropy
from tensorflow.keras.metrics import CategoricalAccuracy, AUC, Precision, Recall
from tensorflow_addons.metrics import F1Score


def get_command_line_arguments():
    parser = argparse.ArgumentParser(description='Train COVID-19 detection model from scratch.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('data', type=str, help='path to the dataset')
    parser.add_argument('model', type=str,
                        help='path where to save trained model, learning curves, checkpoints and logs. Must be a '
                             'non-existing directory.')
    parser.add_argument('--architecture', type=str, default='resnet50',
                        help='architecture to use. Supported: resnet50, covidnet.')
    parser.add_argument('--class-weights', action='store_true', default=False,
                        help='compensate dataset imbalance using class weights')
    parser.add_argument('--data-augmentation', action='store_true', default=False, help='augment data during training')
    parser.add_argument('--load-weights', type=str, default=None,
                        help='path to weights to be loaded (useful for resuming training)')
    parser.add_argument('--initial-epoch', type=int, default=0,
                        help='initial epochs to skip (useful for resuming training)')
    parser.add_argument('--epochs', type=int, default=30, help='epochs of training')
    parser.add_argument('--learning-rate', type=float, default=1e-4, help='learning rate for training')
    return parser.parse_args()


def get_loss():
    return CategoricalCrossentropy()


def get_metrics(dataset_info):
    # remarks:
    # - AUC and F1-score are computed with macro-average (we care a lot about the small COVID-19 class!)
    # - precision and recall are computed only on the COVID-19 class (again, it is the most important)
    n_classes = dataset_info['n_classes']
    metrics = [
        CategoricalAccuracy(name='accuracy'),
        AUC(name='auc', multi_label=True),      # multi_label=True => macro-average
        F1Score(name='f1-score', num_classes=n_classes, average='macro')
    ]
    if 'covid-19' in dataset_info['class_labels']:
        covid19_label = dataset_info['class_labels']['covid-19']
        metrics.append(Precision(name='precision_covid19', class_id=covid19_label))
        metrics.append(Recall(name='recall_covid19', class_id=covid19_label))
    return metrics


def main():
    # command-line arguments
    args = get_command_line_arguments()

    # prepare paths
    dataset_path = Path(args.data)
    models_path = Path(args.model)
    if models_path.is_dir():
        raise FileExistsError(str(models_path) + ' already exists')
    logs_path = models_path / 'logs'
    checkpoints_path = models_path / 'checkpoints'
    plots_path = models_path / 'training'
    models_path = models_path / 'models'
    models_path.mkdir(parents=True)
    checkpoints_path.mkdir()
    plots_path.mkdir()

    # build input pipeline
    train_ds, train_ds_info = image_dataset_from_directory(dataset_path / 'train', IMAGE_SIZE,
                                                           augmentation=args.data_augmentation)
    val_ds, _ = image_dataset_from_directory(dataset_path / 'validation', IMAGE_SIZE, shuffle=False)

    # prepare training stuff
    loss = get_loss()
    metrics = get_metrics(train_ds_info)
    callbacks = get_callbacks(checkpoints_path, logs_path)
    class_weights = get_class_weights(train_ds_info) if args.class_weights else None

    # train whole model from scratch
    model = get_model(args.architecture, None, train_ds_info, args.load_weights)
    history = model.compile_and_fit(args.learning_rate, loss, metrics, train_ds, val_ds, args.epochs,
                                    args.initial_epoch, callbacks, class_weights)
    model.save_weights(str(models_path / 'model'))
    plot_learning_curves(history, save_path=plots_path)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import argparse
import pandas as pd
import matplotlib.pyplot as plt
from pathlib import Path


def get_command_line_arguments():
    parser = argparse.ArgumentParser(description='Plot TensorBoard logs.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('logs', type=str, help='path to directory containing logs')
    parser.add_argument('metric', type=str, help='metric to plot (e.g. accuracy)')
    return parser.parse_args()


def main():
    args = get_command_line_arguments()

    root_path = Path(args.logs)
    lines = [
        (root_path / 'run-train-tag-epoch_accuracy.csv', 'training'),
        (root_path / 'run-validation-tag-epoch_accuracy.csv', 'validation'),
    ]

    plt.figure()
    for line in lines:
        csv_content = pd.read_csv(line[0], engine='python', sep=',')
        x = csv_content['Step']
        y = csv_content['Value']
        plt.plot(x, y, label=line[1])
    plt.legend()
    plt.xlabel('epoch')
    plt.ylabel(args.metric)
    plt.show()


if __name__ == '__main__':
    main()



from tensorflow.keras.layers import Layer, Conv2D, DepthwiseConv2D


class PEPXBlock(Layer):
    """
    Design pattern heavily used in COVID-Net, proposed by Linda Wangg et al. (see docs/COVID-Net), consisting of:
    - First-stage Projection: 1x1 convolutions for projecting input features to a lower dimension.
    - Expansion: 1x1 convolutions for expanding features to a higher dimension.
    - Depth-wise Representation: 3x3 depth-wise convolutions for learning efficiently spatial characteristics.
    - Second-stage Projection: 1x1 convolutions for projecting features back to a lower dimension.
    - Extension: 1x1 convolutions that finally extend channel dimensionality to a higher dimension.

    Since the paper does not provide all the details, some choices have been taken according to the state of the art:
    - No batch normalization and no activation.
    - Number of filters in projections set to 1/2 of the initial number of channels.
    - Number of filters in expansion set to 3/4 of the initial number of channels.
    """

    def __init__(self, channels, **kwargs):
        super().__init__(kwargs)
        self.channels = channels

        nf_p = channels // 2            # number of filters for projections 1 and 2
        nf_e = int(3 / 4 * channels)    # number of filters for expansion

        self._projection_1 = Conv2D(nf_p, 1, name='projection_1')
        self._expansion = Conv2D(nf_e, 1, name='expansion')
        self._depth_wise = DepthwiseConv2D(3, padding='same', name='depth_wise')
        self._projection_2 = Conv2D(nf_p, 1, name='projection_2')
        self._extension = Conv2D(channels, 1, name='extension')

    def call(self, inputs, training=None, mask=None):
        x = self._projection_1(inputs, training=training)
        x = self._expansion(x, training=training)
        x = self._depth_wise(x, training=training)
        x = self._projection_2(x, training=training)
        x = self._extension(x, training=training)
        return x

    def get_config(self):
        config = super().get_config()
        config.update({'channels': self.channels})
        return config



from covid19.models._model import Model
from covid19.layers import Rescaling
from tensorflow.keras import Sequential, Input
from tensorflow.keras.applications import ResNet50V2
from tensorflow.keras.layers import Dense, GlobalAveragePooling2D


class ResNet50(Model):
    """
    COVID-19 detection model with ResNet50v2 as convolutional base for feature extraction.

    Inputs: batches of images with shape (None, 224, 224, 3).
    Outputs: batches of softmax activations (None, n_classes).
    """
    _image_shape = (224, 224, 3)

    def __init__(self, n_classes, name='resnet50', weights='imagenet'):
        """
        :param n_classes: int, number of classes (units in the last layer)
        :param name: string, name of the model
        :param weights: one of 'imagenet', or path to the pretrained weights to load
        """
        super().__init__(name=name)
        self._n_classes = n_classes
        self._transfer_learning = weights is not None

        if weights is not None and weights != 'imagenet':   # weights for the whole model
            load_whole = True
            weights_resnet = None
        else:
            load_whole = False
            weights_resnet = weights

        self._feature_extractor = Sequential([
            Rescaling(1./127.5, offset=-1),
            ResNet50V2(include_top=False, weights=weights_resnet, input_shape=self.image_shape)
        ], name='feature_extractor')

        self._classifier = Sequential([
            GlobalAveragePooling2D(),
            Dense(n_classes, activation='softmax')
        ], name='classifier')

        if load_whole:
            self.load_weights(weights)

        # required for summary()
        inputs = Input(shape=self.image_shape)
        outputs = self.call(inputs)
        super().__init__(name=name, inputs=inputs, outputs=outputs)
        self.build(input_shape=(None, self.image_shape[0], self.image_shape[1], self.image_shape[2]))

    def get_config(self):
        config = super().get_config()
        config.update({'n_classes': self._n_classes})
        return config

    def fit_linear_classifier(self, learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks,
                              class_weights=None):
        self.feature_extractor.trainable = False
        return self.compile_and_fit(learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks,
                                    class_weights)

    def fine_tune(self, learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks, fine_tune_at,
                  class_weights=None):
        if fine_tune_at > len(self.feature_extractor.layers[-1].layers):
            raise ValueError('Too big fine_tune_at, more than the number of layers')
        self.feature_extractor.trainable = True     # unfreeze convolutional base
        for layer in self.feature_extractor.layers[-1].layers[:fine_tune_at]:
            layer.trainable = False                 # freeze bottom layers
        return self.compile_and_fit(learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks,
                                    class_weights)

    @property
    def feature_extractor(self):
        return self._feature_extractor

    @property
    def classifier(self):
        return self._classifier

    @property
    def image_shape(self):
        return self._image_shape

    @property
    def transfer_learning(self):
        return self._transfer_learning



#!/usr/bin/env python3

import argparse
import numpy as np
import tensorflow as tf
from utils import get_model, VERBOSE, IMAGE_SIZE
from tqdm import tqdm
from pathlib import Path
from covid19.datasets import image_dataset_from_directory
from covid19.metrics import plot_confusion_matrix, plot_roc, make_classification_report
from covid19.explainers import GradCAM, IG, plot_explanation


def get_command_line_arguments():
    parser = argparse.ArgumentParser(description='Test COVID-19 detection model.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('analysis', type=str, help='type of evaluation. Supported: performance, explainability.')
    parser.add_argument('data', type=str, help='path to the dataset')
    parser.add_argument('output', type=str, help='path where to save the results')
    parser.add_argument('model', type=str, help='path to the model/checkpoint to test')
    parser.add_argument('architecture', type=str, help='architecture of the model. Supported: resnet50, covidnet.')
    parser.add_argument('--explainer', type=str, default='gradcam', help='Explainer to use. Supported: gradcam, ig')
    return parser.parse_args()


def evaluate(model, dataset, dataset_info, output_path):
    probabilities = model.predict(dataset, verbose=VERBOSE)
    predictions = np.argmax(probabilities, axis=1)
    labels = dataset_info['labels']
    labels_one_hot = tf.one_hot(labels, dataset_info['n_classes'])

    class_names = sorted(dataset_info['class_labels'].keys())
    covid19_label = dataset_info['class_labels']['covid-19']
    covid19_probabilities = probabilities[:, covid19_label]
    covid19_binary_labels = labels_one_hot[:, covid19_label]

    plot_confusion_matrix(labels, predictions, class_names, save_path=output_path)
    plot_roc(covid19_binary_labels, covid19_probabilities, save_path=output_path)
    make_classification_report(labels, predictions, class_names, save_path=output_path)


def explain(model, dataset, dataset_info, output_path, explainer):
    if explainer == 'gradcam':
        explainer = GradCAM(model)
    elif explainer == 'ig':
        explainer = IG(model)
    else:
        raise ValueError('Invalid explainer')

    count = {label: 0 for label in dataset_info['class_labels'].values()}
    class_names = {label: class_name for class_name, label in dataset_info['class_labels'].items()}    # reverse
    n_batches = dataset_info['n_batches']
    iter_dataset = iter(dataset)

    with tqdm(total=n_batches * dataset_info['batch_size']) as bar:
        bar.set_description('Explaining images')

        for _ in range(n_batches):   # can't directly iterate over dataset, as it iterates forever
            batch = next(iter_dataset)
            images = batch[0]
            labels = batch[1]

            for image, label in zip(images, labels):
                bar.update()

                prediction, confidence, explanation = explainer.explain(image)
                label = np.argmax(label)
                count[label] += 1
                save_path = output_path / (class_names[label] + '_{:05d}'.format(count[label]) + '.png')
                plot_explanation(image, explanation, class_names[prediction], confidence, class_names[label],
                                 save_path=save_path)


def main():
    # command-line arguments
    args = get_command_line_arguments()

    # prepare paths
    dataset_path = Path(args.data) / 'test'
    output_path = Path(args.output)
    output_path.mkdir(parents=True, exist_ok=True)

    # build input pipeline
    test_ds, test_ds_info = image_dataset_from_directory(dataset_path, IMAGE_SIZE, shuffle=False)

    # evaluate
    model = get_model(args.architecture, None, test_ds_info, args.model)
    if args.analysis == 'performance':
        evaluate(model, test_ds, test_ds_info, output_path)
    elif args.analysis == 'explainability':
        explain(model, test_ds, test_ds_info, output_path, args.explainer)
    else:
        raise ValueError('Invalid analysis.')


if __name__ == '__main__':
    main()



import numpy as np
import shutil
from tqdm import tqdm
from pathlib import Path


def _get_labels(dataset):
    return {sample[2] for sample in dataset}


def _split_class(dataset, label, split):
    # select samples and IDs for the class
    train_set = [sample for sample in dataset if sample[2] == label]
    test_set = []
    ids = list({sample[0] for sample in train_set})

    # sample IDs until the number of images is enough
    # note: IDs are sampled, not directly samples, since we want all the samples with an ID in the same set
    n_desired = split * len(train_set)
    n = 0
    while n < n_desired:
        id_ = ids.pop(np.random.choice(len(ids)))
        id_samples = {sample for sample in train_set if sample[0] == id_}
        train_set = [sample for sample in train_set if sample not in id_samples]   # remove used samples
        test_set += id_samples
        n += len(id_samples)
    return train_set, test_set


def _copy_or_move_images(dataset, output_path, move=False):
    # create paths
    labels = _get_labels(dataset)
    output_path = Path(output_path)
    output_path.mkdir(exist_ok=True)
    for label in labels:
        label_path = output_path / label
        label_path.mkdir()

    # copy or move images
    with tqdm(total=len(dataset)) as bar:
        bar.set_description('Filling ' + str(output_path))
        for sample in dataset:
            bar.update()
            label = sample[2]
            filepath = sample[1]
            new_filepath = output_path / label / filepath.name
            if move:
                shutil.move(filepath, new_filepath)
            else:
                shutil.copy(filepath, new_filepath)


def stratified_sampling(dataset, split):
    """
    Split a dataset in training set and test set using stratified sampling.

    :param dataset: list of tuples (id, filepath, label)
    :param split: float, percentage of dataset for test set (must be between 0 and 1)
    :return: (train_set, test_set)
    """
    labels = _get_labels(dataset)
    train_set = []
    test_set = []
    for label in labels:
        train_set_, test_set_ = _split_class(dataset, label, split)
        train_set += train_set_
        test_set += test_set_
    return train_set, test_set


def copy_images(dataset, output_path):
    """
    Copies the images of a dataset to a new path.

    :param dataset: list of tuples (id, filepath, label)
    :param output_path: path where to copy the images (destination path).
    """
    _copy_or_move_images(dataset, output_path)


def move_images(dataset, output_path):
    """
    Moves the images of a dataset to a new path.

    :param dataset: list of tuples (id, filepath, label)
    :param output_path: path where to copy the images (destination path).
    """
    _copy_or_move_images(dataset, output_path, move=True)



import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path


def plot_learning_curves(history, history_ft=None, save_path=None):
    """
    Plots the learning curves for all the metrics contained in history (one figure for each metric).

    :param history: History object returned by tf.keras.Model.fit()
    :param history_ft: History object returned by tf.keras.Model.fit(), including the fine-tuning epochs. If this
        argument is provided, history must contain the History object without the fine-tuning epochs.
    :param save_path: path to the directory where to save the figures (with names '<metric>.png')
    """
    if len(history.epoch) == 0:     # empty history, use only history_ft
        epochs = 0
        history = history_ft
        history_ft = None
    else:
        epochs = history.epoch[-1]

    for metric in history.history.keys():
        if metric.startswith('val_'):
            continue                # already taken into account as dual metric of another one

        val_metric = 'val_' + metric
        train_values = history.history[metric]
        val_values = history.history[val_metric]
        if history_ft is not None and metric in history_ft.history:
            train_values += history_ft.history[metric]
            val_values += history_ft.history[val_metric]

        x = np.arange(len(train_values))
        plt.figure()
        plt.plot(x, train_values, label='training')
        plt.plot(x, val_values, label='validation')
        if history_ft is not None:
            plt.plot([epochs, epochs], plt.ylim(), label='start fine-tuning')
        plt.xlabel('epoch')
        plt.ylabel(metric)
        plt.legend()
        if save_path is not None:
            save_path = Path(save_path)
            plt.savefig(save_path / (metric + '.png'))
        plt.close()



#!/usr/bin/env python3

import argparse
from pathlib import Path
from covid19.datasets import generate_covidx, generate_ham10000


def get_command_line_arguments():
    parser = argparse.ArgumentParser(description='Generate dataset and split it in train, validation and test sets.')
    parser.add_argument('name', type=str, help='name of the dataset. Supported: covidx, ham10000.')
    parser.add_argument('data', type=str, help='path to the source datasets')
    parser.add_argument('output', type=str, help='path where to store the dataset')
    return parser.parse_args()


def main():
    # command-line arguments
    args = get_command_line_arguments()

    # prepare paths
    dataset_path = Path(args.data)
    output_path = Path(args.output)
    if not dataset_path.is_dir():
        raise ValueError('Invalid dataset path')
    if output_path.exists():
        raise FileExistsError(str(output_path) + ' already exists')
    output_path.mkdir(parents=True)

    # generate dataset
    if args.name == 'covidx':
        generate_covidx(dataset_path, output_path, seed=1)
    elif args.name == 'ham10000':
        generate_ham10000(dataset_path, output_path, seed=1)
    else:
        raise ValueError('Invalid dataset name.')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import argparse
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path


def get_command_line_arguments():
    parser = argparse.ArgumentParser(description='Examine dataset.')
    parser.add_argument('data', type=str, help='path to the dataset')
    parser.add_argument('output', type=str, help='path where to save the dataset stats')
    return parser.parse_args()


def get_stats(dataset_path):
    train_path = dataset_path / 'train'
    val_path = dataset_path / 'validation'
    test_path = dataset_path / 'test'

    train_class_paths = [path for path in train_path.iterdir()]
    val_class_paths = [path for path in val_path.iterdir()]
    test_class_paths = [path for path in test_path.iterdir()]

    train_class_sizes = [len(list(path.iterdir())) for path in train_class_paths]
    val_class_sizes = [len(list(path.iterdir())) for path in val_class_paths]
    test_class_sizes = [len(list(path.iterdir())) for path in test_class_paths]

    train_total = sum(train_class_sizes)
    val_total = sum(val_class_sizes)
    test_total = sum(test_class_sizes)

    stats = {
        'training': (train_class_sizes, train_total),
        'validation': (val_class_sizes, val_total),
        'test': (test_class_sizes, test_total),
    }
    class_names = [path.name for path in train_class_paths]

    return stats, class_names


def write_stats(stats, class_names, save_path):
    with open(save_path / 'stats.txt', 'w') as f:
        for dataset, dataset_stats in stats.items():
            f.write('{:<15}\n'.format(dataset + ' set:'))
            print('{:<15}'.format(dataset + ' set:'))
            for class_name, class_size in zip(class_names, dataset_stats[0]):
                f.write('{:<15}{:^15}\n'.format(class_name, class_size))
                print('{:<15}{:^15}'.format(class_name, class_size))
            f.write('{:<15}{:^15}\n\n'.format('total', dataset_stats[1]))
            print('{:<15}{:^15}\n'.format('total', dataset_stats[1]))


def plot_stats(stats, class_names, save_path):
    x = np.arange(len(class_names))
    plt.figure()
    plt.bar(x, stats['training'][0], width=0.25, label='training')
    plt.bar(x + 0.25, stats['validation'][0], width=0.25, tick_label=class_names, label='validation')
    plt.bar(x + 0.5, stats['test'][0], width=0.25, label='test')
    plt.ylabel('number of images ($log_{10}$)')
    plt.yscale('log')
    plt.legend(loc='upper left')
    plt.savefig(save_path / 'stats.png')
    plt.show()


def main():
    # command-line arguments
    args = get_command_line_arguments()

    # prepare paths
    dataset_path = Path(args.data)
    output_path = Path(args.output)
    output_path.mkdir(parents=True, exist_ok=True)

    # analyze
    stats, class_names = get_stats(dataset_path)
    write_stats(stats, class_names, output_path)
    plot_stats(stats, class_names, output_path)


if __name__ == '__main__':
    main()



from tensorflow.keras.layers import Layer


class Rescaling(Layer):
    """
    Pre-processing layer to rescale inputs.

    The behavior is identical to tf.keras.layers.experimental.preprocessing.Rescaling, but the latter does not support
    the offset in tensorflow 2.2.
    """

    def __init__(self, scale, offset=0, **kwargs):
        super().__init__(kwargs)
        self.scale = scale
        self.offset = offset

    def call(self, inputs, training=None, mask=None):
        return inputs * self.scale + self.offset

    def get_config(self):
        config = super().get_config()
        config.update({'scale': self.scale, 'offset': self.offset})
        return config



import tensorflow as tf
from abc import ABC, abstractmethod
from tensorflow.keras.optimizers import Adam


class Model(tf.keras.Model, ABC):
    """
    Abstract Base Class for COVID-19 detection models.

    The subclassing models have to contain only two major blocks of layers:
    - feature extractor: convolutional base pre-processing inputs and extracting features. It must end with a
        convolutional layer (required by explainers in covid19.explainers).
    - classifier: classification model based on dense layers, on top of the convolutional base.
    """

    def __init__(self, **kwargs):
        super().__init__(kwargs)

    def compile_and_fit(self, learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks,
                        class_weights):
        """
        Compiles and fits the model.

        :param learning_rate: float, learning rate
        :param loss: tf.keras.Loss, loss function for the optimizer
        :param metrics: list of tf.keras.Metric, metrics of interest
        :param train_ds: tf.data.Dataset, training dataset
        :param val_ds: tf.data.Dataset, validation dataset
        :param epochs: int, number of epochs (effective, not counting the initial epochs to skip)
        :param initial_epoch: int, number of initial epochs to skip
        :param callbacks: list of tf.keras.Callback, callbacks for the training loop
        :param class_weights: dictionary (label -> weight), class weights to compensate dataset imbalance.
        :return: History object
        """
        self.compile(optimizer=Adam(lr=learning_rate), loss=loss, metrics=metrics)
        self.summary()
        return self.fit(train_ds, epochs=epochs+initial_epoch, initial_epoch=initial_epoch, validation_data=val_ds,
                        callbacks=callbacks, class_weight=class_weights)

    def call(self, inputs, training=None, mask=None):
        """Forward pass."""
        # using Rescaling layer, tf.data.Dataset and tf.keras.Model.fit() causes unknown shape... reshaping fixes
        # see https://gitmemory.com/issue/tensorflow/tensorflow/24520/511633717
        x = tf.reshape(inputs, tf.constant((-1,) + self.image_shape))

        # if we are using pretrained weights (self.transfer_learning=True), BN layers must be kept in inference mode
        # see https://www.tensorflow.org/tutorials/images/transfer_learning
        training = False if self.transfer_learning else training

        x = self.feature_extractor(x, training=training)
        x = self.classifier(x, training=training)
        return x

    def get_config(self):
        return super().get_config()

    @abstractmethod
    def fit_linear_classifier(self, learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks,
                              class_weights=None):
        """
        Fits layer on top, freezing the rest of the network. It has to be used as first step of transfer learning,
        before fine-tuning.

        :param learning_rate: float, learning rate
        :param loss: tf.keras.Loss, loss function for the optimizer
        :param metrics: list of tf.keras.Metric, metrics of interest
        :param train_ds: tf.data.Dataset, training dataset
        :param val_ds: tf.data.Dataset, validation dataset
        :param epochs: int, number of epochs (effective, not counting the initial epochs to skip)
        :param initial_epoch: int, number of initial epochs to skip
        :param callbacks: list of tf.keras.Callback, callbacks for the training loop
        :param class_weights: dictionary (label -> weight), class weights to compensate dataset imbalance.
        :return: History object
        """
        raise NotImplementedError

    @abstractmethod
    def fine_tune(self, learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks, fine_tune_at,
                  class_weights=None):
        """
        Fine-tunes some layers. The learning rate should be lower than the normal fit.

        :param learning_rate: float, learning rate
        :param loss: tf.keras.Loss, loss function for the optimizer
        :param metrics: list of tf.keras.Metric, metrics of interest
        :param train_ds: tf.data.Dataset, training dataset
        :param val_ds: tf.data.Dataset, validation dataset
        :param epochs: int, number of epochs (effective, not counting the initial epochs to skip)
        :param initial_epoch: int, number of initial epochs to skip
        :param callbacks: list of tf.keras.Callback, callbacks for the training loop
        :param fine_tune_at: int, index of layer at which to start to unfreeze
        :param class_weights: dictionary (label -> weight), class weights to compensate dataset imbalance.
        :return: History object
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def feature_extractor(self):
        """Convolutional model for feature extraction (e.g. tf.keras.applications.ResNet50). Must end with a
        convolutional layer."""
        raise NotImplementedError

    @property
    @abstractmethod
    def classifier(self):
        """Classification model on top of the convolutional base (e.g. Flatten -> Dense)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def image_shape(self):
        """Shape of the input images (height, width, channels). For example, (224, 224, 3)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def transfer_learning(self):
        raise NotImplementedError



import numpy as np
import pandas as pd
from pathlib import Path
from tqdm import tqdm
from covid19.datasets._utils import stratified_sampling, copy_images


def _process_dataset(dataset_path):
    dataset = []
    image_paths = [dataset_path / 'HAM10000_images_part_1', dataset_path / 'HAM10000_images_part_2']
    metadata_path = dataset_path / 'HAM10000_metadata.csv'
    csv_content = pd.read_csv(metadata_path)

    with tqdm(total=len(csv_content)) as bar:
        bar.set_description('Processing dataset')
        for _, row in csv_content.iterrows():
            bar.update()

            # get fields
            lesion_id = row['lesion_id']
            image_id = row['image_id']
            label = row['dx']

            # find filepath
            filepath_found = False
            for image_path in image_paths:
                filepath = image_path / image_id
                filepath = filepath.with_suffix('.jpg')
                if filepath.is_file():
                    filepath_found = True
                    break
            if not filepath_found:
                raise FileNotFoundError('Image ' + image_id + ' not found')

            # add to dataset
            dataset.append((lesion_id, filepath, label))
    return dataset


def generate_ham10000(dataset_path, output_path, test_split=.15, validation_split=.15, seed=None):
    """
    Generates HAM10000 dataset with the following structure:
    - train
        - class_1
        - class_2
        - ...
    - validation
        - class_1
        - class_2
        - ...
    - test
        - class_1
        - class_2
        - ...

    The source data must be downloaded from https://www.kaggle.com/kmader/skin-cancer-mnist-ham10000 and put in the
    directory indicated by dataset_path.

    :param dataset_path: path to the directory containing the source data
    :param output_path: path where to put HAM10000 dataset
    :param test_split: float, fraction of data to be used as test set (must be between 0 and 1)
    :param validation_split: float, fraction of training data to be used as validation set (must be between 0 and 1)
    :param seed: seed for random number generator
    """
    if seed is not None:
        np.random.seed(seed)

    dataset_path = Path(dataset_path)
    output_path = Path(output_path)

    dataset = _process_dataset(dataset_path)
    train_set, test_set = stratified_sampling(dataset, test_split)
    train_set, val_set = stratified_sampling(train_set, validation_split)

    copy_images(train_set, output_path / 'train')
    copy_images(val_set, output_path / 'validation')
    copy_images(test_set, output_path / 'test')



#!/usr/bin/env python3

import argparse
from pathlib import Path
from utils import get_model, IMAGE_SIZE
from covid19.datasets import image_dataset_from_directory
from tensorflow.keras.layers import Dense


def get_command_line_arguments():
    parser = argparse.ArgumentParser(description='Convert a model to a pretrained version for another dataset.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('data_pretraining', type=str, help='path to the dataset used for pretraining')
    parser.add_argument('data_new', type=str, help='path to the new dataset')
    parser.add_argument('output', type=str, help='path where to save the pretrained model')
    parser.add_argument('model', type=str, help='path to the model to convert.')
    parser.add_argument('architecture', type=str, help='architecture of the model. Supported: resnet50, covidnet.')
    return parser.parse_args()


def main():
    # command-line arguments
    args = get_command_line_arguments()

    # prepare paths
    dataset_pretraining_path = Path(args.data_pretraining) / 'train'
    dataset_new_path = Path(args.data_new) / 'train'
    output_path = Path(args.output) / 'model_pretrained'
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # load model
    _, train_ds_info = image_dataset_from_directory(dataset_pretraining_path, IMAGE_SIZE)
    model = get_model(args.architecture, None, train_ds_info, args.model)

    # replace last layer with the right number of units
    _, train_ds_info = image_dataset_from_directory(dataset_new_path, IMAGE_SIZE)
    model.classifier.pop()
    model.classifier.add(Dense(train_ds_info['n_classes']))

    # save pretrained model
    model.save_weights(output_path)


if __name__ == '__main__':
    main()



from covid19.models._model import Model
from covid19.layers import Rescaling, PEPXBlock
from tensorflow.keras import Sequential, Input
from tensorflow.keras.layers import Layer, Conv2D, BatchNormalization, ReLU, MaxPool2D, Flatten, Dense, add


class _COVIDNetBlock(Layer):
    def __init__(self, channels, n_pepx, **kwargs):
        super().__init__(kwargs)
        self.channels = channels
        self.n_pepx = n_pepx

        self._branch_conv = Conv2D(channels, 1)
        self._branch_pepx = []
        for _ in range(n_pepx):
            self._branch_pepx.append(PEPXBlock(channels))

    def call(self, inputs, training=None, mask=None):
        branch_conv_output = self._branch_conv(inputs, training=training)

        x = inputs
        for i in range(len(self._branch_pepx)):
            x = self._branch_pepx[i](x, training=training)
            x = add([x, branch_conv_output])
        return x

    def get_config(self):
        config = super().get_config()
        config.update({'channels': self.channels, 'n_pepx': self.n_pepx})
        return config


class COVIDNet(Model):
    """
    COVID-19 detection model with specific architecture proposed by Linda Wangg et al. (see docs/COVID-Net).

    Inputs: batches of images with shape (None, 224, 224, 3).
    Outputs: batches of softmax activations (None, 3). The 3 classes are meant to be: covid-19, normal, pneumonia.

    Since the paper does not provide all the details, some choices have been taken according to the state of the art:
    - Batch normalization, ReLU activation, and L2 regularization for the fully connected layers.
    - Max pooling 3x3 with stride 2 at the end of each block (where the spatial dimensionality decreases).
    - No activation and batch normalization for the convolutional layers.
    - Inputs rescaled to the range [-1,1].
    """
    _image_shape = (224, 224, 3)

    def __init__(self, n_classes, name='covidnet', weights=None):
        """
        :param n_classes: int, number of classes (units in the last layer)
        :param name: string, name of the model
        :param weights: path to the pretrained weights to load
        """
        super().__init__(name=name)
        self._n_classes = n_classes
        self._transfer_learning = weights is not None

        self._feature_extractor = Sequential([
            Rescaling(1./127.5, offset=-1),
            Conv2D(64, 7, strides=(2, 2), padding='same'),
            MaxPool2D(pool_size=(3, 3), strides=(2, 2), padding='same'),
            _COVIDNetBlock(256, 3),
            MaxPool2D(pool_size=(3, 3), strides=(2, 2), padding='same'),
            _COVIDNetBlock(512, 4),
            MaxPool2D(pool_size=(3, 3), strides=(2, 2), padding='same'),
            _COVIDNetBlock(1024, 5),
            MaxPool2D(pool_size=(3, 3), strides=(2, 2), padding='same'),
            _COVIDNetBlock(2048, 3),
        ], name='feature_extractor')

        self._classifier = Sequential([
            Flatten(),
            Dense(1024, kernel_regularizer='l2'),
            BatchNormalization(),
            ReLU(),
            Dense(1024, kernel_regularizer='l2'),
            BatchNormalization(),
            ReLU(),
            Dense(n_classes, kernel_regularizer='l2', activation='softmax')
        ], name='classifier')

        if weights is not None:
            self.load_weights(weights)

        # required for summary()
        inputs = Input(shape=self.image_shape)
        outputs = self.call(inputs)
        super().__init__(name=name, inputs=inputs, outputs=outputs)
        self.build(input_shape=(None, self.image_shape[0], self.image_shape[1], self.image_shape[2]))

    def get_config(self):
        config = super().get_config()
        config.update({'n_classes': self._n_classes})
        return config

    def fit_linear_classifier(self, learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks,
                              class_weights=None):
        self.feature_extractor.trainable = False
        self.classifier.trainable = True
        for layer in self.classifier.layers[:-1]:
            layer.trainable = False
        return self.compile_and_fit(learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks,
                                    class_weights)

    def fine_tune(self, learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks, fine_tune_at,
                  class_weights=None):
        n_layers_feature_extractor = len(self.feature_extractor.layers) - 1     # -1 for rescaling layer
        n_layers_classifier = len(self.classifier.layers) - 1                   # -1 for linear classifier on top

        if fine_tune_at > n_layers_feature_extractor + n_layers_classifier:
            raise ValueError('Too big fine_tune_at, more than the number of layers')

        if fine_tune_at > n_layers_feature_extractor:
            # freeze all the convolutional base + part of the classifier
            self.feature_extractor.trainable = False
            self.classifier.trainable = True
            fine_tune_at -= n_layers_feature_extractor
            for layer in self.classifier.layers[:fine_tune_at]:
                layer.trainable = False
        else:
            # freeze part of the convolutional base
            self.feature_extractor.trainable = True
            self._classifier.trainable = True
            for layer in self.feature_extractor.layers[:fine_tune_at+1]:        # +1 for rescaling layer
                layer.trainable = False
            for layer in self.classifier.layers:
                layer.trainable = True

        return self.compile_and_fit(learning_rate, loss, metrics, train_ds, val_ds, epochs, initial_epoch, callbacks,
                                    class_weights)

    @property
    def feature_extractor(self):
        return self._feature_extractor

    @property
    def classifier(self):
        return self._classifier

    @property
    def image_shape(self):
        return self._image_shape

    @property
    def transfer_learning(self):
        return self._transfer_learning



import numpy as np
from covid19.models import ResNet50, COVIDNet
from tensorflow.keras.callbacks import ModelCheckpoint, TensorBoard

VERBOSE = 2
IMAGE_SIZE = (224, 224)


def get_model(architecture, weights, dataset_info, load_weights):
    n_classes = dataset_info['n_classes']
    if architecture == 'resnet50':
        model = ResNet50(n_classes, weights=weights)
    elif architecture == 'covidnet':
        model = COVIDNet(n_classes, weights=weights)
    else:
        raise ValueError('Invalid architecture')
    if load_weights is not None:
        model.load_weights(load_weights)
    return model


def get_callbacks(checkpoints_path, logs_path):
    filepath_checkpoint = checkpoints_path / 'epoch_{epoch:02d}'
    return [
        ModelCheckpoint(filepath=str(filepath_checkpoint), save_weights_only=True, verbose=VERBOSE),
        TensorBoard(log_dir=logs_path, profile_batch=0)
    ]


def get_class_weights(train_ds_info):
    total = train_ds_info['n_images']
    n_classes = train_ds_info['n_classes']
    class_weights = {}

    for class_name, class_label in train_ds_info['class_labels'].items():
        # scale weights by total / n_classes to keep the loss to a similar magnitude
        # see https://www.tensorflow.org/tutorials/structured_data/imbalanced_data#class_weights
        n = len(np.where(train_ds_info['labels'] == class_label)[0])
        class_weights[class_label] = (1 / n) * (total / n_classes)
    return class_weights



from abc import ABC, abstractmethod


class Explainer(ABC):
    """Abstract Base Class for COVID-19 explainers."""

    @abstractmethod
    def explain(self, image):
        """Explains the image by superimposing a heatmap.

        :param image: image to predict and explain
        :return: (prediction, confidence, explanation), where explanation is the superimposed image"""
        raise NotImplementedError



import numpy as np
import tensorflow as tf
from scipy import ndimage
from covid19.explainers._explainer import Explainer
from tensorflow.keras.applications.resnet_v2 import preprocess_input


def _apply_polarity(attributions, polarity):
    if polarity == "positive":
        return np.clip(attributions, 0, 1)
    else:
        return np.clip(attributions, -1, 0)


def _binarize(attributions, threshold=0.001):
    return attributions > threshold


def _get_thresholded_attributions(attributions, percentage):
    if percentage == 100.0:
        return np.min(attributions)

    # 1. Flatten the attributions
    flatten_attr = attributions.flatten()

    # 2. Get the sum of the attributions
    total = np.sum(flatten_attr)

    # 3. Sort the attributions from largest to smallest.
    sorted_attributions = np.sort(np.abs(flatten_attr))[::-1]

    # 4. Calculate the percentage of the total sum that each attribution
    # and the values about it contribute.
    cum_sum = 100.0 * np.cumsum(sorted_attributions) / total

    # 5. Threshold the attributions by the percentage
    indices_to_consider = np.where(cum_sum >= percentage)[0][0]

    # 6. Select the desired attributions and return
    attributions = sorted_attributions[indices_to_consider]
    return attributions


def _morphological_cleanup_fn(attributions, structure=np.ones((4, 4))):
    closed = ndimage.grey_closing(attributions, structure=structure)
    opened = ndimage.grey_opening(closed, structure=structure)
    return opened


def _apply_linear_transformation(attributions, clip_above_percentile=99.9, clip_below_percentile=70.0,
                                 lower_end=0.2):
    # 1. Get the thresholds
    m = _get_thresholded_attributions(attributions, percentage=100 - clip_above_percentile)
    e = _get_thresholded_attributions(attributions, percentage=100 - clip_below_percentile)

    # 2. Transform the attributions by a linear function f(x) = a*x + b such that
    # f(m) = 1.0 and f(e) = lower_end
    transformed_attributions = (1 - lower_end) * (np.abs(attributions) - e) / (
            m - e
    ) + lower_end

    # 3. Make sure that the sign of transformed attributions is the same as original attributions
    transformed_attributions *= np.sign(attributions)

    # 4. Only keep values that are bigger than the lower_end
    transformed_attributions *= transformed_attributions >= lower_end

    # 5. Clip values and return
    transformed_attributions = np.clip(transformed_attributions, 0.0, 1.0)
    return transformed_attributions


def _draw_outlines(attributions, percentage=90, connected_component_structure=np.ones((3, 3))):
    # 1. Binarize the attributions.
    attributions = _binarize(attributions)

    # 2. Fill the gaps
    attributions = ndimage.binary_fill_holes(attributions)

    # 3. Compute connected components
    connected_components, num_comp = ndimage.measurements.label(
        attributions, structure=connected_component_structure
    )

    # 4. Sum up the attributions for each component
    total = np.sum(attributions[connected_components > 0])
    component_sums = []
    for comp in range(1, num_comp + 1):
        mask = connected_components == comp
        component_sum = np.sum(attributions[mask])
        component_sums.append((component_sum, mask))

    # 5. Compute the percentage of top components to keep
    sorted_sums_and_masks = sorted(component_sums, key=lambda x: x[0], reverse=True)
    sorted_sums = list(zip(*sorted_sums_and_masks))[0]
    cumulative_sorted_sums = np.cumsum(sorted_sums)
    cutoff_threshold = percentage * total / 100
    cutoff_idx = np.where(cumulative_sorted_sums >= cutoff_threshold)[0][0]
    if cutoff_idx > 2:
        cutoff_idx = 2

    # 6. Set the values for the kept components
    border_mask = np.zeros_like(attributions)
    for i in range(cutoff_idx + 1):
        border_mask[sorted_sums_and_masks[i][1]] = 1

    # 7. Make the mask hollow and show only the border
    eroded_mask = ndimage.binary_erosion(border_mask, iterations=1)
    border_mask[eroded_mask] = 0

    # 8. Return the outlined mask
    return border_mask


class IG(Explainer):
    """
    Integrated Gradients (IG).

    Supported models: sub-package covid19.models
    Adapted from: https://keras.io/examples/vision/integrated_gradients/
    """

    def __init__(self, model):
        self.model = model
        self.img_size = model.image_shape
        self.positive_channel = [0, 255, 0]
        self.negative_channel = [255, 0, 0]

    def _get_gradients(self, img_input, top_pred_idx):
        images = tf.cast(img_input, tf.float32)
        with tf.GradientTape() as tape:
            tape.watch(images)
            preds = self.model(images)
            top_class = preds[:, top_pred_idx]
        grads = tape.gradient(top_class, images)
        return grads

    def _get_integrated_gradients(self, img_input, top_pred_idx, baseline=None, num_steps=50):
        # If baseline is not provided, start with a black image
        # having same size as the input image.
        if baseline is None:
            baseline = np.zeros(self.img_size).astype(np.float32)
        else:
            baseline = baseline.astype(np.float32)

        # 1. Do interpolation.
        img_input = img_input.astype(np.float32)
        interpolated_image = [
            baseline + (step / num_steps) * (img_input - baseline)
            for step in range(num_steps + 1)
        ]
        interpolated_image = np.array(interpolated_image).astype(np.float32)

        # 2. Preprocess the interpolated images
        interpolated_image = preprocess_input(interpolated_image)

        # 3. Get the gradients
        grads = []
        for i, img in enumerate(interpolated_image):
            img = tf.expand_dims(img, axis=0)
            grad = self._get_gradients(img, top_pred_idx=top_pred_idx)
            grads.append(grad[0])
        grads = tf.convert_to_tensor(grads, dtype=tf.float32)

        # 4. Approximate the integral using the trapezoidal rule
        grads = (grads[:-1] + grads[1:]) / 2.0
        avg_grads = tf.reduce_mean(grads, axis=0)

        # 5. Calculate integrated gradients and return
        integrated_grads = (img_input - baseline) * avg_grads
        return integrated_grads

    def _random_baseline_integrated_gradients(self, img_input, top_pred_idx, num_steps=50, num_runs=2):

        # 1. List to keep track of Integrated Gradients (IG) for all the images
        integrated_grads = []

        # 2. Get the integrated gradients for all the baselines
        for run in range(num_runs):
            baseline = np.random.random(self.img_size) * 255
            igrads = self._get_integrated_gradients(
                img_input=img_input,
                top_pred_idx=top_pred_idx,
                baseline=baseline,
                num_steps=num_steps,
            )
            integrated_grads.append(igrads)

        # 3. Return the average integrated gradients for the image
        integrated_grads = tf.convert_to_tensor(integrated_grads)
        return tf.reduce_mean(integrated_grads, axis=0)

    def _process_grads(self, image, attributions, polarity="positive", clip_above_percentile=99.9,
                       clip_below_percentile=0, morphological_cleanup=False, structure=np.ones((3, 3)), outlines=False,
                       outlines_component_percentage=90, overlay=True):
        if polarity not in ["positive", "negative"]:
            raise ValueError(
                f""" Allowed polarity values: 'positive' or 'negative'
                                    but provided {polarity}"""
            )
        if clip_above_percentile < 0 or clip_above_percentile > 100:
            raise ValueError("clip_above_percentile must be in [0, 100]")

        if clip_below_percentile < 0 or clip_below_percentile > 100:
            raise ValueError("clip_below_percentile must be in [0, 100]")

        # 1. Apply polarity
        if polarity == "positive":
            attributions = _apply_polarity(attributions, polarity=polarity)
            channel = self.positive_channel
        else:
            attributions = _apply_polarity(attributions, polarity=polarity)
            attributions = np.abs(attributions)
            channel = self.negative_channel

        # 2. Take average over the channels
        attributions = np.average(attributions, axis=2)

        # 3. Apply linear transformation to the attributions
        attributions = _apply_linear_transformation(
            attributions,
            clip_above_percentile=clip_above_percentile,
            clip_below_percentile=clip_below_percentile,
            lower_end=0.0,
        )

        # 4. Cleanup
        if morphological_cleanup:
            attributions = _morphological_cleanup_fn(attributions, structure=structure)
        # 5. Draw the outlines
        if outlines:
            attributions = _draw_outlines(attributions, percentage=outlines_component_percentage)

        # 6. Expand the channel axis and convert to RGB
        attributions = np.expand_dims(attributions, 2) * channel

        # 7.Superimpose on the original image
        if overlay:
            attributions = np.clip((attributions * 0.8 + image), 0, 255)
        return attributions

    def _visualize(self, image, integrated_gradients, polarity="positive", clip_above_percentile=99.9,
                   clip_below_percentile=0, morphological_cleanup=False, structure=np.ones((3, 3)), outlines=False,
                   outlines_component_percentage=90, overlay=True):
        # 1. Make a copy of original image
        image_copy = np.copy(image)

        # 2. Process the integrated gradients
        igrads_attr = self._process_grads(
            image=image_copy,
            attributions=integrated_gradients,
            polarity=polarity,
            clip_above_percentile=clip_above_percentile,
            clip_below_percentile=clip_below_percentile,
            morphological_cleanup=morphological_cleanup,
            structure=structure,
            outlines=outlines,
            outlines_component_percentage=outlines_component_percentage,
            overlay=overlay,
        )
        return igrads_attr.astype(np.uint8)

    def explain(self, image):
        image = np.expand_dims(image, axis=0)
        original = np.copy(image[0]).astype(np.uint8)
        probabilities = self.model.predict(image)
        prediction = tf.argmax(probabilities[0]).numpy()
        confidence = probabilities[0, prediction]
        integrated_gradients = self._random_baseline_integrated_gradients(np.copy(original), top_pred_idx=prediction,
                                                                          num_steps=50, num_runs=2).numpy()

        explanation = self._visualize(
            image=original,
            integrated_gradients=integrated_gradients,
            clip_above_percentile=95,
            clip_below_percentile=28,
            morphological_cleanup=True,
            outlines=True,
        )
        return prediction, confidence, explanation



#!/usr/bin/env python3

import sys
from pathlib import Path
from PySide6.QtWidgets import QApplication
from covid19.ui import Client


def main():
    app = QApplication(sys.argv)
    window = Client(
        n_classes=3,
        models_path=Path(__file__).absolute().parent.parent / 'models',
        class_labels={'covid-19': 0, 'normal': 1, 'pneumonia': 2}
    )
    window.showMaximized()
    sys.exit(app.exec())


if __name__ == '__main__':
    main()



import tensorflow as tf
from tensorflow.keras.preprocessing.image import ImageDataGenerator
from tensorflow.keras.preprocessing.image import random_rotation, random_shift, random_zoom, random_brightness

BATCH_SIZE = 32


def _augment(image):
    image = random_rotation(image, 10, row_axis=0, col_axis=1, channel_axis=2, fill_mode='constant')
    image = random_shift(image, 0.1, 0.1, row_axis=0, col_axis=1, channel_axis=2, fill_mode='constant')
    image = random_brightness(image, (0.9, 1.1))
    image = random_zoom(image, (0.85, 1.15), row_axis=0, col_axis=1, channel_axis=2, fill_mode='constant')
    image = tf.image.random_flip_left_right(image)
    return image


def image_dataset_from_directory(dataset_path, image_size, augmentation=False, shuffle=True):
    """
    Generates a tf.data.Dataset from images in a directory.

    :param dataset_path: path to the dataset. The directory must contain one sub-directory for each class.
    :param image_size: tuple (height, width), image size
    :param augmentation: bool, whether to augment the data (horizontal flip, rotation, shift, brightness, zoom)
    :param shuffle: bool, whether to shuffle
    :return: (dataset, info), where dataset is a tf.data.Dataset and info is a dictionary containing useful information
        (e.g. n_batches).
    """
    # flow from directory
    image_flow = ImageDataGenerator().flow_from_directory(dataset_path, target_size=image_size, batch_size=BATCH_SIZE,
                                                          shuffle=shuffle)
    n_classes = image_flow.num_classes
    n_batches = len(image_flow)

    # tf.data.Dataset
    dataset = tf.data.Dataset.from_generator(lambda: image_flow, output_types=(tf.float32, tf.float32),
                                             output_shapes=((None, image_size[0], image_size[1], 3), (None, n_classes)))
    if augmentation:
        dataset = dataset.unbatch()
        dataset = dataset.map(lambda image, label: (tf.numpy_function(_augment, [image], tf.float32), label),
                              num_parallel_calls=tf.data.experimental.AUTOTUNE)
        dataset = dataset.batch(BATCH_SIZE)
    dataset = dataset.take(n_batches)
    dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)

    # useful info
    info = {
        'n_classes': n_classes,
        'n_images': len(image_flow.classes),
        'n_batches': n_batches,
        'class_labels': image_flow.class_indices,
        'batch_size': BATCH_SIZE,
        'labels': image_flow.classes
    }

    return dataset, info



import seaborn as sns
import matplotlib.pyplot as plt
from pathlib import Path
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.metrics import roc_curve, auc


def plot_confusion_matrix(labels, predictions, class_names, save_path=None):
    """
    Plots the confusion matrix. The confusion matrix is annotated with the absolute counts, while colored according
    to the normalized values by row. In this way, even for imbalanced datasets, the diagonal is highlighted well if the
    predictions are good.

    :param labels: numpy array of shape (n_samples,), ground truth (correct labels)
    :param predictions: numpy array of shape (n_samples,), predictions
    :param class_names: list of class names to use as ticks
    :param save_path: path to the directory where to save the figure (with name 'confusion_matrix.png')
    """
    cm = confusion_matrix(labels, predictions)
    cm_normalized = cm / cm.sum(axis=1).reshape(-1, 1)

    plt.figure()
    sns.heatmap(cm_normalized, annot=cm, fmt="d", xticklabels=class_names, yticklabels=class_names, cmap='Reds')
    plt.ylabel('Ground truth')
    plt.xlabel('Prediction')
    if save_path is not None:
        save_path = Path(save_path)
        plt.savefig(save_path / 'confusion_matrix.png')
    plt.close()


def plot_roc(labels, probabilities, save_path=None):
    """
    Plots the ROC curve for binary classification. In case of multi-class labels, the labels should be binarized
    with the respect to a particular class (e.g. labels[:, class_index]).

    :param labels: numpy array of shape (n_samples,), ground truth (correct labels, 1 or 10 or 1)
    :param probabilities: numpy array of shape (n_samples,), probabilities (e.g. softmax activations)
    :param save_path: path to the directory where to save the figure (with name 'roc.png')
    """
    fpr, tpr, _ = roc_curve(labels, probabilities)
    roc_auc = auc(fpr, tpr)

    plt.figure()
    plt.plot(fpr, tpr, label='ROC curve (area={:.2f})'.format(roc_auc), linewidth=2)
    plt.plot([0, 1], [0, 1], label='random choice', linestyle='--', linewidth=2)
    plt.xlabel('false positive rate')
    plt.ylabel('true positive rate')
    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.grid(True)
    plt.legend()
    if save_path is not None:
        save_path = Path(save_path)
        plt.savefig(save_path / 'roc.png')
    plt.close()


def make_classification_report(labels, predictions, class_names, save_path=None):
    """
    Generates a classification report including: precision, recall, f1-score, accuracy,

    :param labels: numpy array of shape (n_samples,), ground truth (correct labels)
    :param predictions: numpy array of shape (n_samples,), predictions
    :param class_names: list of class names to use as ticks
    :param save_path: path to the directory where to save the report (with name 'classification_report.txt')
    """
    save_path = Path(save_path)
    cr = classification_report(labels, predictions, target_names=class_names)
    print(cr)
    if save_path is not None:
        save_path = Path(save_path)
        with open(save_path / 'classification_report.txt', mode='w') as f:
            f.write(cr)



import shutil
import datetime
import cv2
import numpy as np
import pandas as pd
import pydicom as dicom
from pathlib import Path
from tqdm import tqdm
from covid19.datasets._utils import stratified_sampling, move_images

LABELS = ['covid-19', 'normal', 'pneumonia']
PREFIX = 'COVIDx'


def _get_filename(count):
    return PREFIX + '_{:05d}'.format(count) + '.png'


def _copy_image_with_gray_scale(source, destination):
    image = cv2.imread(str(source), cv2.IMREAD_GRAYSCALE)       # to gray-scale
    cv2.imwrite(str(destination), image)


def _process_dataset_1(dataset, dataset_path, tmp_path):
    image_path = dataset_path / 'images'
    metadata_path = dataset_path / 'metadata.csv'
    valid_views = ["PA", "AP", "AP Supine", "AP erect"]
    csv_content = pd.read_csv(metadata_path)
    count = len(dataset)
    urls = set()

    with tqdm(total=len(csv_content)) as bar:
        bar.set_description('Processing dataset 1')
        for _, row in csv_content.iterrows():
            bar.update()

            # skip invalid views (e.g. lateral)
            if row['view'] not in valid_views:
                continue

            # check label
            label = row['finding'].split(',')[0].lower()      # take the first finding
            if label in LABELS:

                # add URL (dataset 4 checks for overlaps)
                url = row['url']
                urls.add(url)

                # check filepath
                filepath = image_path / row['filename']
                if not filepath.is_file():
                    raise FileNotFoundError('File ' + str(filepath) + ' not found')

                # copy image in gray-scale
                new_filepath = tmp_path / (PREFIX + '_{:05d}'.format(count) + '.png')
                _copy_image_with_gray_scale(filepath, new_filepath)
                count += 1

                # add to dataset
                patient_id = row['patientid']
                dataset.append((patient_id, new_filepath, label))
    return dataset, urls


def _process_dataset_2(dataset, dataset_path, tmp_path):
    image_path = dataset_path / 'images'
    metadata_path = dataset_path / 'metadata.csv'
    csv_content = pd.read_csv(metadata_path, encoding='ISO-8859-1')
    count = len(dataset)

    with tqdm(total=len(csv_content)) as bar:
        bar.set_description('Processing dataset 2')
        for _, row in csv_content.iterrows():
            bar.update()

            # check label
            label = row['finding']
            if str(label) != 'nan':
                label = label.lower()
            if label in LABELS:

                # check filepath
                patient_id = row['patientid']
                filepath = image_path / patient_id
                if filepath.with_suffix('.jpg').is_file():
                    filepath = filepath.with_suffix('.jpg')
                elif filepath.with_suffix('.png').is_file():
                    filepath = filepath.with_suffix('.png')
                else:
                    raise FileNotFoundError('File ' + str(filepath) + ' not found')

                # copy image in gray-scale
                new_filepath = tmp_path / _get_filename(count)
                _copy_image_with_gray_scale(filepath, new_filepath)
                count += 1

                # add to dataset
                dataset.append((patient_id, new_filepath, label))
    return dataset


def _process_dataset_3(dataset, dataset_path, tmp_path):
    image_path = dataset_path / 'images'
    metadata_path = dataset_path / 'metadata.csv'
    csv_content = pd.read_csv(metadata_path)
    count = len(dataset)

    with tqdm(total=len(csv_content)) as bar:
        bar.set_description('Processing dataset 3')
        for _, row in csv_content.iterrows():
            bar.update()

            # check label
            label = row['finding']
            if str(label) != 'nan':
                label = label.lower()
            if label in LABELS:

                # check filepath
                patient_id = row['patientid']
                filepath = image_path / row['imagename']
                if not filepath.is_file():
                    raise FileNotFoundError('File ' + str(filepath) + ' not found')

                # copy image in gray-scale
                new_filepath = tmp_path / _get_filename(count)
                _copy_image_with_gray_scale(filepath, new_filepath)
                count += 1

                # add to dataset
                dataset.append((patient_id, new_filepath, label))
    return dataset


def _process_dataset_4(dataset, dataset_path, tmp_path, urls):
    image_path = dataset_path / 'COVID-19'
    metadata_path = dataset_path / 'COVID-19.metadata.xlsx'
    bad_patient_ids = {'100', '101', '102', '103', '104', '105', '110', '111', '112', '113', '122', '123', '124', '125',
                       '126', '217'}
    csv_content = pd.read_excel(metadata_path)
    count = len(dataset)

    with tqdm(total=len(csv_content)) as bar:
        bar.set_description('Processing dataset 4')
        for _, row in csv_content.iterrows():
            bar.update()
            label = LABELS[0]

            # skip bad images
            patient_id = row['FILE NAME'].split('(')[1].split(')')[0]
            if patient_id in bad_patient_ids:
                continue

            # check URL (overlaps)
            url = row['URL']
            if url in urls:
                continue

            # check filepath
            suffix = '.' + row['FORMAT'].lower()
            filepath = image_path / row['FILE NAME']
            filepath = filepath.with_suffix(suffix)
            if not filepath.is_file():
                filename = row['FILE NAME'].split('(')[0] + ' (' + row['FILE NAME'].split('(')[1]
                filepath = filepath.with_name(filename).with_suffix(suffix)
                if not filepath.is_file():
                    raise FileNotFoundError('File ' + str(filepath) + ' not found')

            # copy image in gray-scale
            new_filepath = tmp_path / _get_filename(count)
            _copy_image_with_gray_scale(filepath, new_filepath)
            count += 1

            # add to dataset
            dataset.append((patient_id, new_filepath, label))
    return dataset


def _process_dataset_5_sample(dataset, file_paths, tmp_path, row, image_path, label):
    patient_id = row['patientId']
    filepath = image_path / patient_id
    filepath = filepath.with_suffix('.dcm')
    count = len(dataset)

    # check filepath
    if filepath not in file_paths:  # repetition
        if not filepath.is_file():
            raise FileNotFoundError('File ' + str(filepath) + ' not found')
        file_paths.add(filepath)

        # copy image from .dcm to .png
        ds = dicom.dcmread(filepath)
        pixel_array_numpy = ds.pixel_array
        new_filepath = tmp_path / _get_filename(count)
        cv2.imwrite(str(new_filepath), pixel_array_numpy)

        # copy image in gray-scale
        _copy_image_with_gray_scale(new_filepath, new_filepath)

        # add to dataset
        dataset.append((patient_id, new_filepath, label))
    return dataset, file_paths


def _process_dataset_5(dataset, dataset_path, tmp_path):
    image_path = dataset_path / 'stage_2_train_images'
    file_paths = set()

    # get normal samples
    metadata_path = dataset_path / 'stage_2_detailed_class_info.csv'
    csv_content = pd.read_csv(metadata_path)
    with tqdm(total=len(csv_content)) as bar:
        bar.set_description('Processing dataset 5 (part 1)')
        for _, row in csv_content.iterrows():
            bar.update()
            label = row['class'].lower()
            if label in LABELS:
                dataset, file_paths = _process_dataset_5_sample(dataset, file_paths, tmp_path, row, image_path, label)

    # get pneumonia samples
    metadata_path = dataset_path / 'stage_2_train_labels.csv'
    csv_content = pd.read_csv(metadata_path)
    with tqdm(total=len(csv_content)) as bar:
        bar.set_description('Processing dataset 5 (part 2)')
        for _, row in csv_content.iterrows():
            bar.update()
            label = row['Target']
            if label == 1:
                label = LABELS[2]
                dataset, file_paths = _process_dataset_5_sample(dataset, file_paths, tmp_path, row, image_path, label)
    return dataset


def generate_covidx(dataset_path, output_path, test_split=.15, validation_split=.15, seed=None):
    """
    Generates COVIDx dataset using the following sources:
    - 1) https://github.com/ieee8023/covid-chestxray-dataset
    - 2) https://github.com/agchung/Figure1-COVID-chestxray-dataset
    - 3) https://github.com/agchung/Actualmed-COVID-chestxray-dataset
    - 4) https://www.kaggle.com/tawsifurrahman/covid19-radiography-database
    - 5) https://www.kaggle.com/c/rsna-pneumonia-detection-challenge
    These datasets must be downloaded and put in the directory indicated by dataset_path.

    The generated COVIDx dataset is put in the directory indicated by output_path with the following structure:
    - train
        - covid-19
        - normal
        - pneumonia
    - validation
        - covid-19
        - normal
        - pneumonia
    - test
        - covid-19
        - normal
        - pneumonia

    Adapted from: https://github.com/lindawangg/COVID-Net/blob/master/create_COVIDx.ipynb

    :param dataset_path: path to the directory containing the datasets described above
    :param output_path: path where to put COVIDx dataset
    :param test_split: float, fraction of data to be used as test set (must be between 0 and 1)
    :param validation_split: float, fraction of training data to be used as validation set (must be between 0 and 1)
    :param seed: seed for random number generator
    """
    if seed is not None:
        np.random.seed(seed)

    dataset_path = Path(dataset_path)
    output_path = Path(output_path)
    tmp_path = output_path / ('tmp' + datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
    shutil.rmtree(tmp_path, ignore_errors=True)
    tmp_path.mkdir()

    dataset = []
    dataset, urls = _process_dataset_1(dataset, dataset_path / 'covid-chestxray-dataset', tmp_path)
    dataset = _process_dataset_2(dataset, dataset_path / 'Figure1-COVID-chestxray-dataset', tmp_path)
    dataset = _process_dataset_3(dataset, dataset_path / 'Actualmed-COVID-chestxray-dataset', tmp_path)
    dataset = _process_dataset_4(dataset, dataset_path / 'COVID-19 Radiography Database', tmp_path, urls)
    dataset = _process_dataset_5(dataset, dataset_path / 'rsna-pneumonia-detection-challenge', tmp_path)

    train_set, test_set = stratified_sampling(dataset, test_split)
    train_set, val_set = stratified_sampling(train_set, validation_split)

    move_images(train_set, output_path / 'train')
    move_images(val_set, output_path / 'validation')
    move_images(test_set, output_path / 'test')
    shutil.rmtree(tmp_path)



from pathlib import Path
from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()
with open('requirements.txt') as f:
    install_requires = f.read()
scripts = [str(script) for script in list(Path('scripts').iterdir())]

setup(
    name='covid19-detection',
    version='0.2.2',
    description='Detection of COVID-19 from Chest X-Ray Images',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Franco Ruggeri, Fredrik Danielsson, Muhammad Tousif Zaman, Milan Jolić',
    author_email='fruggeri@kth.se, fdaniels@kth.se, mtzaman@kth.se, jolic@kth.se',
    license='GPL',
    license_file='LICENSE',
    packages=find_packages(include=['covid19', 'covid19.*']),
    include_package_data=False,
    scripts=scripts,
    url='https://github.com/franco-ruggeri/dd2424-covid19-detection',
    install_requires=install_requires
)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'client.ui'
##
## Created by: Qt User Interface Compiler version 6.1.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

import covid19.ui._resource_rc

class Ui_Client(object):
    def setupUi(self, Client):
        if not Client.objectName():
            Client.setObjectName(u"Client")
        Client.resize(1450, 1049)
        icon = QIcon()
        icon.addFile(u":/images/logo.jpeg", QSize(), QIcon.Normal, QIcon.Off)
        Client.setWindowIcon(icon)
        Client.setStyleSheet(u"* {\n"
"	font-size: 18px;\n"
"}\n"
"\n"
"QPushButton {\n"
"	border-radius: 10px;\n"
"	padding: 5px;\n"
"}\n"
"\n"
"QPushButton:pressed {\n"
"	border-radius: 12px;\n"
"}\n"
"\n"
"QPushButton#predict {\n"
"	background-color: rgb(239, 41, 41);\n"
"	color: white;\n"
"}\n"
"\n"
"QPushButton#predict:hover {\n"
"	background-color: rgb(164, 0, 0);\n"
"}\n"
"\n"
"QPushButton#predict:disabled {\n"
"	background-color: rgb(197, 114, 114)\n"
"}\n"
"\n"
"QPushButton#select_image {\n"
"	background-color: rgb(75, 116, 22);\n"
"	color: white;\n"
"}\n"
"\n"
"QPushButton#select_image:hover {\n"
"	background-color: rgb(22, 57, 9);\n"
"}\n"
"\n"
"QPushButton#select_image:disabled {\n"
"	background-color: rgba(75, 116, 22, 50%)\n"
"}\n"
"\n"
"QGroupBox * {\n"
"	margin: 10px;\n"
"	padding: 5px;\n"
"}")
        self.centralwidget = QWidget(Client)
        self.centralwidget.setObjectName(u"centralwidget")
        self.gridLayout = QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.output_panel = QVBoxLayout()
        self.output_panel.setObjectName(u"output_panel")
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.explanation_label = QLabel(self.centralwidget)
        self.explanation_label.setObjectName(u"explanation_label")
        self.explanation_label.setAlignment(Qt.AlignCenter)

        self.verticalLayout_2.addWidget(self.explanation_label)

        self.explanation = QLabel(self.centralwidget)
        self.explanation.setObjectName(u"explanation")
        self.explanation.setMinimumSize(QSize(640, 640))
        self.explanation.setMaximumSize(QSize(640, 640))
        self.explanation.setPixmap(QPixmap(u":/images/default.png"))
        self.explanation.setScaledContents(True)

        self.verticalLayout_2.addWidget(self.explanation)


        self.output_panel.addLayout(self.verticalLayout_2)

        self.results = QGroupBox(self.centralwidget)
        self.results.setObjectName(u"results")
        self.formLayout_2 = QFormLayout(self.results)
        self.formLayout_2.setObjectName(u"formLayout_2")
        self.prediction_label = QLabel(self.results)
        self.prediction_label.setObjectName(u"prediction_label")

        self.formLayout_2.setWidget(0, QFormLayout.LabelRole, self.prediction_label)

        self.prediction = QLabel(self.results)
        self.prediction.setObjectName(u"prediction")

        self.formLayout_2.setWidget(0, QFormLayout.FieldRole, self.prediction)

        self.confidence_label = QLabel(self.results)
        self.confidence_label.setObjectName(u"confidence_label")

        self.formLayout_2.setWidget(3, QFormLayout.LabelRole, self.confidence_label)

        self.confidence = QLabel(self.results)
        self.confidence.setObjectName(u"confidence")

        self.formLayout_2.setWidget(3, QFormLayout.FieldRole, self.confidence)


        self.output_panel.addWidget(self.results)


        self.gridLayout.addLayout(self.output_panel, 0, 3, 1, 1)

        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer_3, 0, 4, 1, 1)

        self.input_panel = QVBoxLayout()
        self.input_panel.setObjectName(u"input_panel")
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.input_label = QLabel(self.centralwidget)
        self.input_label.setObjectName(u"input_label")
        self.input_label.setAlignment(Qt.AlignCenter)

        self.verticalLayout.addWidget(self.input_label)

        self.input = QLabel(self.centralwidget)
        self.input.setObjectName(u"input")
        self.input.setMinimumSize(QSize(640, 640))
        self.input.setMaximumSize(QSize(640, 640))
        self.input.setPixmap(QPixmap(u":/images/default.png"))
        self.input.setScaledContents(True)

        self.verticalLayout.addWidget(self.input)


        self.input_panel.addLayout(self.verticalLayout)

        self.settings = QGroupBox(self.centralwidget)
        self.settings.setObjectName(u"settings")
        self.formLayout = QFormLayout(self.settings)
        self.formLayout.setObjectName(u"formLayout")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.model_state = QLabel(self.settings)
        self.model_state.setObjectName(u"model_state")
        self.model_state.setMinimumSize(QSize(50, 50))
        self.model_state.setMaximumSize(QSize(50, 50))
        self.model_state.setPixmap(QPixmap(u":/images/error.png"))
        self.model_state.setScaledContents(True)

        self.horizontalLayout_2.addWidget(self.model_state)

        self.model_label = QLabel(self.settings)
        self.model_label.setObjectName(u"model_label")
        self.model_label.setMinimumSize(QSize(0, 50))
        self.model_label.setMaximumSize(QSize(16777215, 50))

        self.horizontalLayout_2.addWidget(self.model_label)


        self.formLayout.setLayout(5, QFormLayout.LabelRole, self.horizontalLayout_2)

        self.model = QComboBox(self.settings)
        self.model.addItem("")
        self.model.addItem("")
        self.model.setObjectName(u"model")
        self.model.setMinimumSize(QSize(0, 50))
        self.model.setMaximumSize(QSize(250, 50))

        self.formLayout.setWidget(5, QFormLayout.FieldRole, self.model)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.explainer_state = QLabel(self.settings)
        self.explainer_state.setObjectName(u"explainer_state")
        self.explainer_state.setMinimumSize(QSize(50, 50))
        self.explainer_state.setMaximumSize(QSize(50, 50))
        self.explainer_state.setPixmap(QPixmap(u":/images/error.png"))
        self.explainer_state.setScaledContents(True)

        self.horizontalLayout_3.addWidget(self.explainer_state)

        self.explainaer_label = QLabel(self.settings)
        self.explainaer_label.setObjectName(u"explainaer_label")
        self.explainaer_label.setMinimumSize(QSize(0, 50))
        self.explainaer_label.setMaximumSize(QSize(16777215, 50))

        self.horizontalLayout_3.addWidget(self.explainaer_label)


        self.formLayout.setLayout(6, QFormLayout.LabelRole, self.horizontalLayout_3)

        self.explainer = QComboBox(self.settings)
        self.explainer.addItem("")
        self.explainer.addItem("")
        self.explainer.setObjectName(u"explainer")
        self.explainer.setMinimumSize(QSize(0, 50))
        self.explainer.setMaximumSize(QSize(250, 50))

        self.formLayout.setWidget(6, QFormLayout.FieldRole, self.explainer)

        self.predict = QPushButton(self.settings)
        self.predict.setObjectName(u"predict")
        self.predict.setEnabled(True)
        self.predict.setMinimumSize(QSize(0, 50))
        self.predict.setMaximumSize(QSize(200, 50))

        self.formLayout.setWidget(8, QFormLayout.FieldRole, self.predict)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.input_state = QLabel(self.settings)
        self.input_state.setObjectName(u"input_state")
        self.input_state.setMinimumSize(QSize(50, 50))
        self.input_state.setMaximumSize(QSize(50, 50))
        self.input_state.setPixmap(QPixmap(u":/images/error.png"))
        self.input_state.setScaledContents(True)

        self.horizontalLayout.addWidget(self.input_state)

        self.select_image_label = QLabel(self.settings)
        self.select_image_label.setObjectName(u"select_image_label")
        self.select_image_label.setMinimumSize(QSize(0, 50))
        self.select_image_label.setMaximumSize(QSize(16777215, 50))

        self.horizontalLayout.addWidget(self.select_image_label)


        self.formLayout.setLayout(7, QFormLayout.LabelRole, self.horizontalLayout)

        self.select_image = QPushButton(self.settings)
        self.select_image.setObjectName(u"select_image")
        self.select_image.setEnabled(True)
        self.select_image.setMinimumSize(QSize(0, 50))
        self.select_image.setMaximumSize(QSize(200, 50))

        self.formLayout.setWidget(7, QFormLayout.FieldRole, self.select_image)


        self.input_panel.addWidget(self.settings)


        self.gridLayout.addLayout(self.input_panel, 0, 1, 1, 1)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer, 0, 0, 1, 1)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer_2, 0, 2, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 1, 1, 1, 1)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer_2, 1, 3, 1, 1)

        Client.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(Client)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 1450, 26))
        Client.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(Client)
        self.statusbar.setObjectName(u"statusbar")
        Client.setStatusBar(self.statusbar)

        self.retranslateUi(Client)

        QMetaObject.connectSlotsByName(Client)
    # setupUi

    def retranslateUi(self, Client):
        Client.setWindowTitle(QCoreApplication.translate("Client", u"COVID-19 Detector", None))
        self.explanation_label.setText(QCoreApplication.translate("Client", u"Explanation", None))
        self.explanation.setText("")
        self.results.setTitle(QCoreApplication.translate("Client", u"Results", None))
        self.prediction_label.setText(QCoreApplication.translate("Client", u"Prediction:", None))
        self.prediction.setText("")
        self.confidence_label.setText(QCoreApplication.translate("Client", u"Confidence:", None))
        self.confidence.setText("")
        self.input_label.setText(QCoreApplication.translate("Client", u"Chest X-Ray Image", None))
        self.input.setText("")
        self.settings.setTitle(QCoreApplication.translate("Client", u"Settings", None))
        self.model_state.setText("")
        self.model_label.setText(QCoreApplication.translate("Client", u"Model:", None))
        self.model.setItemText(0, QCoreApplication.translate("Client", u"ResNet50", None))
        self.model.setItemText(1, QCoreApplication.translate("Client", u"COVID-Net", None))

        self.explainer_state.setText("")
        self.explainaer_label.setText(QCoreApplication.translate("Client", u"Explainer:", None))
        self.explainer.setItemText(0, QCoreApplication.translate("Client", u"Grad-CAM", None))
        self.explainer.setItemText(1, QCoreApplication.translate("Client", u"Integrated Gradients", None))

        self.predict.setText(QCoreApplication.translate("Client", u"Predict", None))
        self.input_state.setText("")
        self.select_image_label.setText(QCoreApplication.translate("Client", u"Input image:", None))
        self.select_image.setText(QCoreApplication.translate("Client", u"Select image", None))
    # retranslateUi




import numpy as np
import matplotlib.cm as cm
import tensorflow as tf
from covid19.explainers._explainer import Explainer
from tensorflow.keras.preprocessing.image import array_to_img, img_to_array


class GradCAM(Explainer):
    """
    Grad-CAM (Gradient-weighted Class Activation Mapping).

    Supported models: sub-package covid19.models
    Adapted from: https://keras.io/examples/vision/grad_cam/
    """

    def __init__(self, model):
        self.model = model

    def _predict(self, image):
        image = tf.expand_dims(image, axis=0)  # add batch dimension

        with tf.GradientTape() as tape:
            last_conv_activations = self.model.feature_extractor(image)
            tape.watch(last_conv_activations)
            probabilities = self.model.classifier(last_conv_activations)
            prediction = tf.argmax(probabilities[0]).numpy()
            top_probability = probabilities[:, prediction]
            confidence = top_probability[0]

        gradients = tape.gradient(top_probability, last_conv_activations)
        return prediction, confidence, last_conv_activations, gradients

    @staticmethod
    def _make_heatmap(last_conv_activations, gradients):
        pooled_grads = tf.reduce_mean(gradients, axis=(0, 1, 2))
        last_conv_activations = last_conv_activations.numpy()[0]
        pooled_grads = pooled_grads.numpy()
        for i in range(pooled_grads.shape[-1]):
            last_conv_activations[:, :, i] *= pooled_grads[i]

        heatmap = np.mean(last_conv_activations, axis=-1)
        heatmap = np.maximum(heatmap, 0) / np.max(heatmap)              # range [0, 1]
        return heatmap

    def explain(self, image):
        # predict
        prediction, confidence, last_conv_activations, gradients = self._predict(image)

        # compute heatmap
        heatmap = self._make_heatmap(last_conv_activations, gradients)  # range [0, 1]
        heatmap = np.uint8(255 * heatmap)                               # range [0, 255] (for colormap)

        # colorize new image according to heatmap (with colormap 'jet')
        jet = cm.get_cmap("jet")
        jet_colors = jet(np.arange(256))[:, :3]
        jet_heatmap = jet_colors[heatmap]
        jet_heatmap = array_to_img(jet_heatmap)
        jet_heatmap = jet_heatmap.resize((image.shape[1], image.shape[0]))
        jet_heatmap = img_to_array(jet_heatmap)

        # superimposed image
        superimposed_image = jet_heatmap * 0.4 + image
        superimposed_image /= np.max(superimposed_image)    # range [0, 1] (for plt.imshow())
        return prediction, confidence, superimposed_image



import matplotlib.pyplot as plt


def plot_explanation(original, explanation, prediction, confidence, ground_truth, save_path=None):
    """
    Plots the original image alongside the explanation.

    :param original: numpy array of shape (height, width, channels), original image
    :param explanation: numpy array of shape (height, width, channels), heatmap superimposed to original image
    :param prediction: string, name of predicted class
    :param confidence: float, probability of predicted class
    :param ground_truth: string, name of true class
    :param save_path: path where to save the figure
    :return:
    """
    plt.figure()

    plt.subplot(1, 2, 1)
    plt.imshow(original / 255)
    plt.axis('off')
    plt.title('original')

    plt.subplot(1, 2, 2)
    plt.imshow(explanation)
    plt.axis('off')
    plt.title('explanation')
    plt.suptitle('Prediction: {} (Confidence: {:.2f}%)\nGround truth: {}'
                 .format(prediction, confidence*100, ground_truth))

    if save_path is not None:
        plt.savefig(save_path)
    plt.close()
